# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: dddsimport.py,v 1.5 2005/06/07 07:43:03 mbaas Exp $

import os.path
from . import _core
from .cgtypes import *
from .globalscene import getScene
from .worldobject import WorldObject
from .trimesh import TriMesh
from .trimeshgeom import TriMeshGeom
from .targetcamera import TargetCamera
from .glpointlight import GLPointLight
from .gltargetspotlight import GLTargetSpotLight
from .spotlight3ds import SpotLight3DS
from .glmaterial import GLMaterial
from .material3ds import Material3DS, TextureMap3DS
from . import pluginmanager
from .sl import *

TYPE_UNKNOWN = 0
TYPE_AMBIENT = 1
TYPE_OBJECT = 2
TYPE_CAMERA = 3
TYPE_TARGET = 4
TYPE_LIGHT = 5
TYPE_SPOT = 6

GEOM_INIT_NORMALS   = 0x01
GEOM_INIT_MATIDS    = 0x02
GEOM_INIT_TEXELS    = 0x04
GEOM_INIT_FLAGS     = 0x08
GEOM_INIT_SMOOTHING = 0x10
GEOM_INIT_ALL       = 0xff


# DDDSImporter
class DDDSImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["3ds"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "3D Studio"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename, flags=GEOM_INIT_ALL, parent=None):
        """Import a 3DS file."""

        self.filename = filename
        self.ddds = _core.File3ds()
        self.ddds.load(filename)
#        f = self.ddds.current_frame
        f = getScene().timer().frame
        if f<self.ddds.segment_from:
            f = self.ddds.segment_from
        if f>self.ddds.segment_to:
            f = self.ddds.segment_to
        self.ddds.eval(f)

        # Create a dictionary containing all available meshes.
        # Key is the mesh name. This is used to check if all meshes have
        # been processed (i.e. if there were corresponding nodes).
        self.meshes = {}
        m = self.ddds.meshes()
        while m!=None:
            if m.name in self.meshes:
                print("Warning: Duplicate mesh names in 3ds file")
            self.meshes[m.name] = m
            m = next(m)

        # Create objects...
        self.materials = {}
        self.createObject(self.ddds.nodes(), parent=parent, flags=flags)
        del self.materials

        # Create TriMeshes for all left over meshes...
        for n in self.meshes:
            mesh = self.meshes[n]
            tm = TriMeshGeom()
            mesh.initGeom(tm, flags)
            worldobj = TriMesh(name = mesh.name, parent=parent)
            worldobj.geom = tm

        self.ddds.free()

        
    def createObject(self, node, parent=None, flags=GEOM_INIT_ALL):
        """
        \param node (\c Lib3dsNode) Current node
        \param parent (\c WorldObject) Parent object or None
        \param flags (\c int) Flags for mesh creation

        \todo worldtransform als Slot
        \todo Pivot so setzen wie in 3DS
        """
        if node==None:
            return

        worldobj = None
        auto_insert = True
        if parent!=None:
            auto_insert = False

        # Object?
        if node.type==TYPE_OBJECT:
            if node.name!="$$$DUMMY" and node.name!="_Quader01":
                data = node.object_data
                mesh = self.ddds.meshByName(node.name)
                if mesh.name in self.meshes:
                    del self.meshes[mesh.name]
#                print "###",node.name,"###"
#                print "Node matrix:"
#                print node.matrix
#                print "Pivot:",data.pivot
#                print "Mesh matrix:"
#                print mesh.matrix
                tm = TriMeshGeom()
                mesh.initGeom(tm, flags)

                if parent==None:
                    PT = mat4().translation(-data.pivot)
                    m = node.matrix*PT*mesh.matrix.inverse()
                else:
                    PT = mat4().translation(-data.pivot)
                    m = node.matrix*PT*mesh.matrix.inverse()
                    # worldtransform von Parent bestimmen...
                    pworldtransform = parent.worldtransform
#                    pworldtransform = parent.localTransform()
#                    p = parent.parent
#                    while p!=None:
#                        pworldtransform = p.localTransform()*pworldtransform
#                        p = p.parent
                    m = pworldtransform.inverse()*m
                worldobj = TriMesh(name=node.name,
#                                   pivot=data.pivot,
                                   transform=m,
                                   auto_insert=auto_insert)
                worldobj.geom = tm

                # Set the materials
                matnames = mesh.materialNames()
                worldobj.setNumMaterials(len(matnames))
                for i in range(len(matnames)):
                    if matnames[i]=="":
                        material = GLMaterial()  
                    # Was the material already instantiated?
                    elif matnames[i] in self.materials:
                        material = self.materials[matnames[i]]
                    else:
                        mat = self.ddds.materialByName(matnames[i])
#                        print "Material:",matnames[i]
#                        print "  self_illum:",mat.self_illum
#                        print "  self_ilpct:",mat.self_ilpct
                        material = Material3DS(name = matnames[i],
                                               ambient = mat.ambient,
                                               diffuse = mat.diffuse,
                                               specular = mat.specular,
                                               shininess = mat.shininess,
                                               shin_strength = mat.shin_strength,
                                               use_blur = mat.use_blur,
                                               transparency = mat.transparency,
                                               falloff = mat.falloff,
                                               additive = mat.additive,
                                               use_falloff = mat.use_falloff,
                                               self_illum = mat.self_illum,
                                               self_ilpct = getattr(mat, "self_ilpct", 0),
                                               shading = mat.shading,
                                               soften = mat.soften,
                                               face_map = mat.face_map,
                                               two_sided = mat.two_sided,
                                               map_decal = mat.map_decal,
                                               use_wire = mat.use_wire,
                                               use_wire_abs = mat.use_wire_abs,
                                               wire_size = mat.wire_size,
                                               texture1_map = self._createTexMap(mat.texture1_map),
                                               texture2_map = self._createTexMap(mat.texture2_map),
                                               opacity_map = self._createTexMap(mat.opacity_map),
                                               bump_map = self._createTexMap(mat.bump_map),
                                               specular_map = self._createTexMap(mat.specular_map),
                                               shininess_map = self._createTexMap(mat.shininess_map),
                                               self_illum_map = self._createTexMap(mat.self_illum_map),
                                               reflection_map = self._createTexMap(mat.reflection_map)
                                               )
                        self.materials[matnames[i]] = material
                    worldobj.setMaterial(material, i)
                
        # Camera?
        elif node.type==TYPE_CAMERA:
            cam = self.ddds.cameraByName(node.name)
#            print "Camera:",node.name
#            print "  Roll:",cam.roll
#            print node.matrix
            # Convert the FOV from horizontal to vertical direction
            # (Todo: What aspect ratio to use?)
            fov = degrees(atan(480/640.0*tan(radians(cam.fov/2.0))))*2.0
            worldobj = TargetCamera(name = node.name,
                                    pos = cam.position,
                                    target = cam.target,
                                    fov = fov,
                                    auto_insert = auto_insert)
        # Light?
        elif node.type==TYPE_LIGHT:
            lgt = self.ddds.lightByName(node.name)
            worldobj = self._createLight(node, lgt, auto_insert)

        if worldobj!=None and parent!=None:
            parent.addChild(worldobj)

        if worldobj!=None:
            self.createObject(node.childs(), worldobj, flags=flags)
        else:
            self.createObject(node.childs(), parent, flags=flags)
        self.createObject(next(node), parent, flags=flags)
        

    ## protected:

    def _createLight(self, node, lgt, auto_insert):
        """Create a light source.

        Helper method for the createObject() method.
        """
        shadow_size = lgt.shadow_size
        shadow_bias = lgt.shadow_bias
        shadow_filter = lgt.shadow_filter
        # Check if a shadow map is used but the parameters are all zero
        # (this happens when the Use Global Settings checkbox in MAX
        # is set). If it happens then use the same default values as MAX
        if lgt.shadowed and shadow_size==0 and shadow_bias==0 and shadow_filter==0:
            shadow_size = 256
            shadow_bias = 3.0
            shadow_filter = 5.0
            
        if lgt.spot_light:
            worldobj = SpotLight3DS(name = node.name,
                                    pos = lgt.position,
                                    target = lgt.spot,
                                    enabled = not lgt.off,
                                    intensity = lgt.multiplier,
                                    see_cone = lgt.see_cone,
                                    roll = lgt.roll, 
                                    outer_range = lgt.outer_range,
                                    inner_range = lgt.inner_range,
                                    attenuation = lgt.attenuation,
                                    rectangular_spot = lgt.rectangular_spot,
                                    shadowed = lgt.shadowed,
                                    shadow_bias = shadow_bias,
                                    shadow_filter = shadow_filter,
                                    shadow_size = shadow_size,
                                    spot_aspect = lgt.spot_aspect,
                                    use_projector = lgt.use_projector,
                                    projector = lgt.projector,
                                    overshoot = lgt.spot_overshoot,
                                    ray_shadows = lgt.ray_shadows,
                                    ray_bias = lgt.ray_bias,
                                    hotspot = lgt.hotspot,
                                    falloff = lgt.falloff,
                                    color = lgt.color,
                                    auto_insert = auto_insert)
#            worldobj = GLTargetSpotLight(name = node.name,
#                                         pos = lgt.position,
#                                         target = lgt.spot,
#                                         cutoff = lgt.falloff/2,
#                                         intensity = lgt.multiplier,
#                                         diffuse = lgt.color,
#                                         auto_insert = auto_insert)
#            if lgt.shadowed:
#                worldobj.shadowmap = (lgt.shadow_size, 1.0, 0.001)
        else:
            worldobj = SpotLight3DS(name = node.name,
                                    pos = lgt.position,
                                    target = lgt.spot,
                                    enabled = not lgt.off,
                                    intensity = lgt.multiplier,
                                    see_cone = lgt.see_cone,
                                    roll = lgt.roll, 
                                    outer_range = lgt.outer_range,
                                    inner_range = lgt.inner_range,
                                    attenuation = lgt.attenuation,
                                    rectangular_spot = lgt.rectangular_spot,
                                    shadowed = False,
                                    shadow_bias = shadow_bias,
                                    shadow_filter = shadow_filter,
                                    shadow_size = shadow_size,
                                    spot_aspect = lgt.spot_aspect,
                                    use_projector = lgt.use_projector,
                                    projector = lgt.projector,
                                    overshoot = True,
                                    ray_shadows = lgt.ray_shadows,
                                    ray_bias = lgt.ray_bias,
                                    hotspot = lgt.hotspot,
                                    falloff = lgt.falloff,
                                    color = lgt.color,
                                    auto_insert = auto_insert)
#            worldobj = GLPointLight(name = node.name,
#                                    pos = lgt.position,
#                                    intensity = lgt.multiplier,
#                                    diffuse = lgt.color,
#                                    auto_insert = auto_insert)
            
        return worldobj
        
    def _createTexMap(self, texmap):
        """Create a TextureMap3DS object (or None).

        \param texmap (\c Lib3dsTextureMap) Texture map
        """
        if texmap.name=="":
            return None

        dir = os.path.dirname(self.filename)
        dir = os.path.abspath(dir)
        return TextureMap3DS(name=os.path.join(dir, texmap.name),
                             flags = texmap.flags,
                             percent = texmap.percent,
                             blur = texmap.blur,
                             scale = texmap.scale,
                             offset = texmap.offset,
                             rotation = texmap.rotation,
                             tint1 = texmap.tint_1,
                             tint2 = texmap.tint_2,
                             tintr = texmap.tint_r,
                             tintg = texmap.tint_g,
                             tintb = texmap.tint_b)

        


######################################################################

# Register the Importer class as a plugin class
if hasattr(_core, "File3ds"):
    pluginmanager.register(DDDSImporter)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: plyimport.py,v 1.4 2005/05/09 14:34:24 mbaas Exp $

import os.path
from . import _core
from .cgtypes import *
from .geomobject import *
from .polyhedron import *
from . import pluginmanager


# PLYImporter
class PLYImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["ply"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "Polygon (PLY)"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename, includevar=None, excludevar=None, invertfaces=False):
        """Import a PLY file.

        includevar is a list of primitive variable names that should be
        imported. excludevar is a list of primitive variable names that
        should not be imported.
        invertfaces specifies if the face orientations should be inverted
        or not.
        """

        self.includevar = includevar
        self.excludevar = excludevar

        imp = _core.PLYReader()
        imp.open(filename)
        # Obtain the header information
        header = imp.readHeader()

        # Create the variable declaration tuples for importing the data...
        inttypes = ["int8", "uint8", "int16", "uint16", "int32", "uint32",
                    "char", "uchar", "short", "ushort", "int", "uint"]
        vardecl = []
        vec3vars = {}
        elements, comment, objinfo = header
        for elname, ninstances, properties in elements:
            for propname, type, len_type, val_type in properties:
                # Vertex indices? (they are handled by default)
                if elname=="vertex" and propname in ["x", "y", "z"]:
                    continue
                # Vertex indices? (they are handled by default)
                if elname=="face" and propname in ["vertex_indices"]:
                    continue
                # Normals?
                if propname in ["nx", "ny", "nz"]:
                    if "N" not in vec3vars and self.isVarAccepted("N"):
                        vardecl.append(("N", NORMAL, elname, ("nx", "ny", "nz")))
                        vec3vars["N"] = True
                    continue

                if type=="list":
                    type = val_type
                if type in inttypes:
                    t = INT
                else:
                    t = FLOAT
                if self.isVarAccepted(propname):
                    vardecl.append((propname, t, elname, (propname,)))

#        print vardecl

        # Create a polyhedron
        name = os.path.splitext(os.path.basename(filename))[0]
        p = Polyhedron(name=name)
        # Set the comment and objinfo
        if comment!="":
            p.geom.newVariable("comment", CONSTANT, STRING)
            s = p.geom.slot("comment")
            s[0] = comment
        if objinfo!="":
            p.geom.newVariable("obj_info", CONSTANT, STRING)
            s = p.geom.slot("obj_info")
            s[0] = objinfo

        # Read the model
        imp.read(p.geom, vardecl, invertfaces)

        imp.close()

    # isVarAccepted
    def isVarAccepted(self, name):
        """Return True if the variable should be imported.

        name is the name of the primitive variable (which is not necessarily
        the ply property!).
        """
        if self.includevar!=None:
            if name not in self.includevar:
                return False
        if self.excludevar!=None:
            if name in self.excludevar:
                return False
        return True

######################################################################

# Register the Importer class as a plugin class
pluginmanager.register(PLYImporter)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: stlimport.py,v 1.1 2005/01/09 20:13:06 mbaas Exp $

import os.path, sys, struct
from .cgtypes import *
from .trimesh import TriMesh
from . import pluginmanager

# STLReader
class STLReader:
    
    def __init__(self, filename):
        self.filename = filename

    # read
    def read(self):
        if self.isASCII():
            self.read_ascii()
        else:
            self.read_bin()

    # read_bin
    def read_bin(self):
        """Read a binary STL file.
        """
        f = file(self.filename, "rb")

        s = f.read(80)
        objname = s.split("\000")[0]
        s = f.read(4)
        numfaces = struct.unpack("<i", s)[0]

        self.begin(objname)

        for i in range(numfaces):
            s = f.read(50)
            normal = vec3(struct.unpack("<fff", s[:12]))
            v = struct.unpack("<fffffffff", s[12:48])
            verts = [vec3(v[0:3]), vec3(v[3:6]), vec3(v[6:9])]
            self.triangle(normal, verts)

        self.end(objname)

    # read_ascii
    def read_ascii(self):
        """Read a ASCII STL file.
        """
        
        f = file(self.filename)
        
        objname = "unnamed"
        normal = vec3()
        verts = []

        state = 0
        linenr = 0
        for s in f:
            linenr+=1
            a = s.split()
            if len(a)==0:
                continue
            
            # Begin of file
            if state==0:
                if a[0]=="solid":
                    if len(a)>1:
                        objname = a[1]
                    self.begin(objname)
                    state = 1
                else:
                    raise SyntaxError('Keyword "solid" expected in line %d'%linenr)
            # Begin of a triangle ("facet") or end of file ("endsolid")
            elif state==1:
                if a[0]=="facet":
                    if len(a)!=5 or a[1]!="normal":
                        raise SyntaxError('Syntax error in line %d'%linenr)
                    normal = vec3(float(a[2]), float(a[3]), float(a[4]))
                    verts = []
                    state = 2
                elif a[0]=="endsolid":
                    self.end(objname)
                    state = 0
                else:
                    raise SyntaxError('Syntax error in line %d. Facet or end of solid expected.'%linenr)
            # "outer loop"
            elif state==2:
                if len(a)!=2 or a[0]!="outer" or a[1]!="loop":
                    raise SyntaxError('Keyword "outer loop" expected in line %d.'%linenr)
                state = 3
            # A vertex definition
            elif state==3:
                if len(a)!=4 or a[0]!="vertex":
                    raise SyntaxError('Syntax error in line %d: Vertex expected'%linenr)
                v = vec3(float(a[1]), float(a[2]), float(a[3]))
                verts.append(v)
                if len(verts)==3:
                    state = 4
            # "endloop"
            elif state==4:
                if a[0]!="endloop":
                    raise SyntaxError('Keyword "endloop" expected in line %d.'%linenr)
                state = 5
            # "endfacet"
            elif state==5:
                if a[0]!="endfacet":
                    raise SyntaxError('Keyword "endfacet" expected in line %d.'%linenr)
                self.triangle(normal, verts)
                state = 1

    # isASCII
    def isASCII(self):
        f = file(self.filename)
        s1 = f.readline().strip()
        s2 = f.readline().strip()
        f.close()
        if s1[:5]!="solid":
            return False
        if s2[:5]!="facet":
            return False
        return True

    # begin
    def begin(self, name):
        """Solid begin callback.

        This callback is called before the triangles are read.
        """
        pass

    # end
    def end(self, name):
        """Solid begin callback.

        This callback is called after the triangles were read.
        """
        pass

    # triangle
    def triangle(self, normal, verts):
        """Triangle callback.

        normal is the normal vector as vec3 and verts is a list with the
        three vertices (vec3s).
        """
        pass


# STLImport
class STLImport(STLReader):
    
    def __init__(self, filename):
        STLReader.__init__(self, filename)
        self.verts = []
        self.numfaces = 0

    # begin
    def begin(self, name):
        pass

    # end
    def end(self, name):
        faces = []
        for i in range(self.numfaces):
            faces.append(list(range(i*3, i*3+3)))
        TriMesh(name=name, verts=self.verts, faces=faces)

    # triangle
    def triangle(self, normal, verts):
        """Triangle callback.

        normal is the normal vector as vec3 and verts is a list with the
        three vertices (vec3s).
        """
        self.numfaces += 1
        self.verts += verts

# STLImporter
class STLImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["stl"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "StereoLithography"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename):
        """Import a STL file."""

        reader = STLImport(filename)
        reader.read()


######################################################################

# Register the Importer class as a plugin class
pluginmanager.register(STLImporter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 18 13:39:48 2019

@author: psakicki
"""

########## BEGIN IMPORT ##########
#### External modules
import numpy as np
import pandas as pd
import datetime as dt
#from pytwobodyorbit import TwoBodyOrbit
#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils
from geodezyx import files_rw
from geodezyx import operational
from geodezyx import time_series

##########  END IMPORT  ##########


def ECI_2_kepler_elts(P,V,rad2deg=True,
                      mu=3.9860044188e14):
    Pnorm = np.linalg.norm(P)
    Vnorm = np.linalg.norm(V)
    
    ### orbital momentum vector H
    H = np.cross(P,V)
    Hnorm = np.linalg.norm(H)
    
    ### eccentricity vector Ecc
    Ecc = (np.cross(V,H)/mu) - (P/Pnorm)
    Ecc_norm = np.linalg.norm(Ecc)
    
    ### true anomaly ν (nu)
    N = np.cross(np.array([0,0,1]),H)
    Nnorm = np.linalg.norm(N)
    if np.dot(P,V) >= 0:
        nu = np.arccos(np.dot(Ecc,P)/(Ecc_norm*Pnorm))
    else:
        nu = 2*np.pi - np.arccos(np.dot(Ecc,P)/(Ecc_norm*Pnorm))        
    
    ### orbit inclination i
    i = np.arccos(H[2]/Hnorm)
    
    ### orbit eccentricity ecc
    ecc = np.linalg.norm(Ecc)
    
    ### eccentric anomaly E (called e here)
    #e1 = 2*np.arctan(np.tan(nu/2)/np.sqrt((1+ecc)/(1-ecc)))
    e2 = 2*np.arctan2(np.tan(nu/2),np.sqrt((1+ecc)/(1-ecc)))
    e=e2
    
    ### longitude of the ascending node Ω (omega_LAN)
    if N[1] >= 0:
        omega_LAN = np.arccos(N[0]/Nnorm)
    else:
        omega_LAN = 2*np.pi - np.arccos(N[0]/Nnorm)
    
    ### argument of periapsis ω (omega_periarg)
    if Ecc[2] >= 0:
        omega_periarg = np.arccos((np.dot(N,Ecc))/(Nnorm * Ecc_norm))
    else:
        omega_periarg = 2*np.pi - np.arccos((np.dot(N,Ecc))/(Nnorm * Ecc_norm))
    
    ### mean anomaly m, (Kepler’s Equation)
    m = e - ecc*np.sin(e)
    
    ### semi-major axis a
    a = ((2/Pnorm) - (Vnorm**2 / mu))**-1

    if rad2deg:
        i=np.rad2deg(i)
        omega_periarg=np.rad2deg(omega_periarg)
        omega_LAN=np.rad2deg(omega_LAN)
        m=np.rad2deg(m)

    return a,ecc,i,omega_periarg,omega_LAN,m



def extrapolate_orbit_kepler(P,V,t,t0,mu=3.9860044188e14):
    orbit = TwoBodyOrbit("", mu=mu)   # create an instance
    orbit.setOrbCart(t0, P, V)        # define the orbit
    Pout, Vout = orbit.posvelatt(t)   # get position and velocity at t1
    kepl = orbit.elmKepl()            # get classical orbital elements
    
    return Pout,Vout,kepl



def extrapolate_sp3_DataFrame(DFsp3,step=900,n_step=9):
    NewEpoch_stk = []
    
    for sat in DFsp3["sat"].unique():
        print(sat)
        DFsat = DFsp3[(DFsp3["sat"] == sat) & (DFsp3["type"] == "P")]
        
        DFline_dummy = DFsat.iloc[0].copy()
        DFline_dummy["clk"] = 999999.999999
        
        XYZ   = DFsat[["x","y","z"]].values *1000
        Tgps = conv.datetime64_numpy2dt(DFsat["epoch"].values)
        Tutc = conv.dt_gpstime2dt_utc(Tgps)
        Tsec = [e.total_seconds() for e in (Tutc-Tutc[0])]
        
        P=conv.ECEF2ECI(XYZ,Tutc)
        V=np.gradient(P,Tsec,axis=0)
        
    
        def extrapo_intern_fct(i_ref,coef):
            ##
            ## backward :
            ## i_ref , coef = 0,-1
            ##
            ## forward :
            ## i_ref , coef = -1,1
            ##
            ## CORRDS ARE GIVEN IN ECI HERE !!!
            ##
            orbit_back = TwoBodyOrbit("orbit", mu=3.9860044188e14)  # create an instance    
            orbit_back.setOrbCart(Tsec[i_ref], P[i_ref], V[i_ref])  # define the orbit
            for t in np.arange(step,n_step*step +1,step):
                Pout, Vout = orbit_back.posvelatt(coef * t)
                epoc = Tgps[i_ref] + coef * dt.timedelta(seconds=int(t))
                DFline_new = DFline_dummy.copy()
                DFline_new[["x","y","z"]] = Pout
                DFline_new["epoch"] = pd.Timestamp(epoc)
        
                NewEpoch_stk.append(DFline_new)
                
            return Pout, Vout
    
        ### backward
        extrapo_intern_fct(0,-1)
    
        ### forward
        extrapo_intern_fct(-1,1)
        
    DFNewEpoch = pd.DataFrame(NewEpoch_stk)
    
    Tutc_NewEpoch = conv.dt_gpstime2dt_utc(conv.datetime64_numpy2dt(DFNewEpoch["epoch"].values))
    DFNewEpoch[["x","y","z"]] = conv.ECI2ECEF(DFNewEpoch[["x","y","z"]].values,Tutc_NewEpoch)
    DFNewEpoch[["x","y","z"]] = DFNewEpoch[["x","y","z"]] * .001
    
    DFout = pd.concat((DFsp3,DFNewEpoch))
    DFout.reset_index(drop=True,inplace=True)
    DFout.sort_values(["sat","epoch"],inplace=True)
    DFout.reset_index(drop=True,inplace=True)
    
    return DFout





# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

## \file keys.py
## Contains the Keys class.

import wx
import string

# Exceptions:
class InvalidKeyDescription(Exception):
    """Exception class."""
    pass

class KeyNotBound(Exception):
    """Exception class."""
    pass

# Keys
class Keys(object):
    """This class manages key strokes.

    Key bindings correspond to special key attributes. The attribute
    name is the name of the key and the value is a callable object
    that takes no arguments.

    Example:

    \code
    >>> keys = Keys()

    # Bind functions to keys...
    >>> keys.ctrl_c = onExit
    keys.shift_ctrl_tab = onFoo

    # Remove a key binding
    >>> del keys.ctrl_c
    
    \endcode
    """

    def __init__(self):

        # Command dictionary. The keys are the tuples that contain the
        # modifier flags and the key code. The value is a callable object.
        self._commands = {}
        
        # A dictionary that translates wx key codes into a string description
        # The dictionary is built from the WXK_xyz constants
        self._wxkeystrs = {}
        wxkeys = [x for x in dir(wx) if x[0:4]=="WXK_"]
        for k in wxkeys:
            code = getattr(wx, k)
            self._wxkeystrs[code] = k[4:].capitalize()

        # The Window object that uses this key manager
        self._window = None


    def attach(self, window):
        self._window = window
        wx.EVT_KEY_DOWN(window, self._onKeyDown)
#        wx.EVT_KEY_UP(window, self._onKeyUp)
#        wx.EVT_CHAR(window, self._onChar)

    def detach(self):
        pass

    def findCommandKeys(self, func):
        """Search the key table for a command and return all associated keys.

        \param func (\c callable) The bound function
        \return A list of readable key strings.
        """

        res = []
        # Compare all commands...
        for key in self._commands.keys():
            cmd = self._commands[key]
            if func==cmd:
                res.append(self._cmdkey2text(key))

        return res

    ######################################################################
    ## protected:

#    def __setitem__(self, name, value):
#        print "setitem"
#        key = self._str2cmdkey(name)
#        print "Setting",key,"to",value
#        self._commands[key] = value        

    def _onKeyDown(self, event):
        """Handle a KeyDown event.

        Note: This event handler is \b not called if the corresponding
        description text is present in the menu. In that case, the menu
        handles the events itself (a wx feature). But it seems this only
        works with English modifier names.
        """
        print(("KeyDown",))
        print(("KeyCode:",event.GetKeyCode(),"RawKeyCode:",event.GetRawKeyCode()))

        # Create the key tuple
        key = self._event2cmdkey(event)
        # Check if the key was bound
        if key in self._commands:
            # Call the bound function
            func = self._commands[key]
            func()
        else:
            event.Skip()
    
    def _onKeyUp(self, event):
#        print "KeyUp",
#        print "KeyCode:",event.GetKeyCode(),"RawKeyCode:",event.GetRawKeyCode()
        event.Skip()

    def __setattr__(self, name, value):
        if name=="":
            return
        # If the name starts with an underscore then it's not a key description
        if name[0]=="_":
            object.__setattr__(self, name, value)
            return

        # Bind key...

        # Create the key tuple
        key = self._str2cmdkey(name)
        # Remove an existing binding
        if key in self._commands:
            self.__delattr__(name)
        # Store the key binding
        self._commands[key] = value

        # Update menu items
        self._updateMenu(value)

    def __delattr__(self, name):
        # If the name starts with an underscore then it's not a key description
        if name[0]=="_":
            object.__delattr__(self, name)
            return

        # Create the key tuple
        key = self._str2cmdkey(name)
        # Check if a key binding exists
        if key in self._commands:
            func = self._commands[key]
            del self._commands[key]
            # Update menu items
            self._updateMenu(func)
        else:
            raise KeyNotBound("Key '%s' is not bound to a function."%name)


    def _updateMenu(self, func):
        """Update all menu items that are bound to func."""
        # Check if any menu items have to be updated
        menu = getattr(self._window, "menu", None)
        if menu==None:
            return
        nodes = menu.findCommandNodes(func)
        for n in nodes:
            n.update()        
        

    def _event2cmdkey(self, event):
        """Convert a key event into a tuple which can be used as key."""
        return (bool(event.ShiftDown()), bool(event.ControlDown()),
                bool(event.AltDown()), bool(event.MetaDown()),
                event.GetKeyCode())

    def _str2cmdkey(self, s):
        """Convert a key description string into the key tuple."""

        f = s.upper().split("_")

        # Check if the modifiers are correct
        for m in f[:-1]:
            if m not in ["SHIFT", "CTRL", "ALT", "META"]:
                raise InvalidKeyDescription("Key '%s' contains invalid modifiers."%s)

        # Convert the last argument into a key code
        skey = f[-1]
        if len(skey)==1:
            key = ord(skey)
        else:
            key = getattr(wx, "WXK_%s"%skey, None)
            if key==None:
                raise InvalidKeyDescription("Key '%s' does not exist."%s)
        
        return ("SHIFT" in f, "CTRL" in f, "ALT" in f, "META" in f, key)

    def _cmdkey2text(self, key):
        shift, ctrl, alt, meta, key = key
        mods = []
        if shift:
            mods.append("Shift")
#            mods.append("Umschalt")
        if ctrl:
            mods.append("Ctrl")
#            mods.append("Strg")
        if alt:
            mods.append("Alt")
        if meta:
            mods.append("Meta")

        ks = self._wxkeystrs.get(key, None)
        if ks==None:
            ks=chr(key)
        mods.append(ks)
            
        return string.join(mods, "+")
        
            
        
    
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: toolbars.py,v 1.1.1.1 2004/12/12 14:31:41 mbaas Exp $

import os.path, string
import wx
from . import panelicons

# ToolBarPalette
class ToolBarPalette(object):
    """Toolbar palette.

    This class manages a wx.Notebook that contains an arbitrary number
    of ToolBars.

    \see ToolBar
    """
    def __init__(self, parent, IDmanager):
        
        self._wx = wx.Notebook(parent, -1, style=wx.CLIP_CHILDREN)
        self.sizer = wx.NotebookSizer(self._wx)

        self._toolbars = []

        self._id_manager = IDmanager

        self.addToolBar("Main")
        self.addToolBar("Test")

        

    def __iter__(self):
        return iter(self._toolbars)

    def __getattr__(self, name):
        return self.findToolBar(name)

        
    def addToolBar(self, title):
        tb = ToolBar(self, title, self._id_manager)
        self._wx.AddPage(tb._panel, title)

        self._toolbars.append(tb)
        
#        tb.add("Open.png", None)
#        tb.add("Close", None)


    # findToolBar
    def findToolBar(self, name):
        """Returns the toolbar with the given name.

        \param name (\c str) Toolbar name
        \return Toolbar (\c ToolBar) or None.
        """
        for tb in self._toolbars:
            if tb.name==name:
                return tb
        return None



        
    ######################################################################
    ## protected:

    # "wx" property...
    
    def _getWx(self):
        """Return the encapsulated wx widget.

        This method is used for retrieving the \a wx property.

        \return wxPython object.
        """
        return self._wx

    wx = property(_getWx, None, None, "Encapsulated wx widget")
        

# ToolBar
class ToolBar:
    """Encapuslates a wx ToolBar.

    \see ToolBarPalette
    """
    
    def __init__(self, parent, name, IDmanager):
        """Constructor.
        
        \param parent (\c ToolBarPalette) Parent tool bar palette 
        """
        self._panel = wx.Panel(parent.wx, -1)
        self._wx = wx.ToolBar(self._panel, -1)
        self._wx.SetName(name)
        self._wx.SetToolBitmapSize(wx.Size(32,32))

        self._sizer = wx.BoxSizer(wx.VERTICAL)
        # Add 8 pixels at the bottom (why is it too small without?!?)
        self._sizer.Add(self._wx, 0, wx.EXPAND | wx.BOTTOM, 8)
        self._panel.SetAutoLayout(True)
        self._panel.SetSizer(self._sizer)

        self._id_manager = IDmanager
        # Lookup table
        self._tool_id_lut = {}

    def __str__(self):
        return "<ToolBar '%s'>"%self._wx.GetName()

    def add(self, bitmapdef, command=None):
        """Add a new tool.
        """

        bitmap = self._def2bitmap(bitmapdef)
        id = self._id_manager.allocID()
        self._wx.AddSimpleTool(id, bitmap, "Short", "A help string.")

        self._tool_id_lut[id] = command
        wx.EVT_MENU(self._wx, id, self._onToolSelection)
        
        self._wx.Realize()


    ######################################################################
    ## protected:

    def _onToolSelection(self, event):
        id = event.GetId()
        if id not in self._tool_id_lut:
            return

        cmd = self._tool_id_lut[id]
        if cmd!=None:
            cmd()

    def _def2bitmap(self, bitmapdef):
        """Convert a bitmap definition into a bitmap.

        \return wx Bitmap
        """
        # Is bitmapdef already a bitmap?
        if isinstance(bitmapdef, wx.BitmapPtr):
            return bitmapdef

        if self._isFilename(bitmapdef):
            bmp = wx.Bitmap(bitmapdef)
            if bmp.Ok():
                return bmp
            else:
                bitmapdef = os.path.basename(bitmapdef)

        bmp = wx.EmptyBitmap(32,32)
        dc = wx.MemoryDC()
        dc.SelectObject(bmp)

        dc.SetPen(wx.BLACK_PEN)
        dc.SetBackground(wx.GREY_BRUSH)
        font = wx.Font(5, wx.SWISS, wx.NORMAL, wx.NORMAL)
        dc.SetFont(font)
        dc.Clear()
        w,h = dc.GetTextExtent(bitmapdef)
        x = (32-w)/2
        y = 32-h-3
        dc.DrawText(bitmapdef,x,y)
        return bmp

    def _isFilename(self, name):
        """Check if a string contains a filename.

        The check is done on the suffix. If there is one that has
        3 or 4 ascii characters then the function returns True.
        """
        n,ext = os.path.splitext(name)
        if ext=="" or len(ext)<4 or len(ext)>5:
            return False
        for c in ext[1:]:
            if c not in string.ascii_letters:
                return False
        return True


    # "wx" property...
    
    def _getWx(self):
        """Return the encapsulated wx widget.

        This method is used for retrieving the \a wx property.

        \return wxPython object.
        """
        return self._wx

    wx = property(_getWx, None, None, "Encapsulated wx widget")

    # "name" property...
    
    def _getName(self):
        """Return the node name.

        This method is used for retrieving the \a name property.

        \return Name (\c str)
        """
        return self._wx.GetName()
        
    name = property(_getName, None, None, "Name")
        




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# -------------------------------------------------------------
# The RenderMan (R) Interface Procedures and Protocol are:
# Copyright 1988, 1989, 2000, Pixar
# All Rights Reserved
#
# RenderMan (R) is a registered trademark of Pixar
# -------------------------------------------------------------
# $Id: slparams.py,v 1.4 2006/02/14 19:29:39 mbaas Exp $

"""Extract shader parameters from a RenderMan shader source file.

Functions:

- slparams()
- convertdefault()
"""

import os, os.path, sys, string, io, types, math
from . import sltokenize
from . import cgtypes, sl, simplecpp
from . import _slparser
from ._slreturntypes import _ShaderInfo, _ShaderParam
import collections
try:
    from . import sloargs
    _has_sloargs = True
except ImportError as exc:
    _has_sloargs = False
    _sloargs_importerror = exc

class SLParamsError(Exception):
    pass

class PreprocessorNotFound(SLParamsError):
    pass

class SyntaxError(SLParamsError):
    pass

class NoMoreTokens(SLParamsError):
    pass


# Parser class (subclassed from the generated Yapps parser)
class _SLParser(_slparser._SLParserBase):
    """SL parser class.

    This class is derived from the generated parser and implements
    some missing methods.
    """
    def __init__(self, scanner):
        _slparser._SLParserBase.__init__(self, scanner)

        # Current filename
        self.filename = "?"
        # Offset which has to be subtracted from the line number to
        # get the true line number within the file.
        self.linenroffset = 0

        # Parameter attributes...
        self.output   = ""
        self.detail   = ""
        self.type     = ""
        self.arraylen = None
        self.name     = ""
        self.space    = None
        self.default  = ""
        self.spaces   = []

        # Shader parameters
        self.params = []

    def newParams(self):
        """Start a new shader.

        The parameter list is cleared.
        """
        self.params = []

    def newType(self):
        """Clear all type parameters.

        This is called when a new type is declared (which is not equivalent
        to a new parameter, e.g. "float a,b,c")
        """
        self.output   = ""
        self.detail   = "uniform"
        self.type     = ""
        self.arraylen = None
        self.name     = ""
        self.space    = None
        self.default  = ""
        self.spaces   = []

    def defaultSpace(self, typ):
        """Return the default space for typ."""
        if typ in ["point", "vector", "normal", "matrix"]:
            return "current"
        elif typ=="color":
            return "rgb"
        else:
            return None
        
    def appendSpace(self):
        """Append self.space to self.spaces."""
        if self.space!=None:
            self.spaces.append(self.space)
        else:
            self.spaces.append(self.defaultSpace(self.type))
                
        self.space = None

    def storeParam(self):
        """Store the current set of attributes as a new parameter.

        The attributes are reset so that a new parameter can begin.
        """
        if self.arraylen==None:
            if self.space==None:
                self.space = self.defaultSpace(self.type)
            self.params.append(_ShaderParam(self.output, self.detail, self.type, None,
                            self.name, self.space, self.default))
        else:
            spaces = self.spaces
            if self.defaultSpace(self.type)==None:
                spaces = None
            self.params.append(_ShaderParam(self.output, self.detail, self.type,
                               self.arraylen, self.name, spaces, self.default))
        self.arraylen = None
        self.name = ""
        self.space = None
        self.default = ""
        self.spaces   = []
        

    def switchFile(self, cppline):
        """Switch to another file.

        This method is called when a preprocessor line (starting with #)
        is read. This line contains the current file name and the line number.
        """
        f = cppline.strip().split(" ")
        linenr = int(f[1])
        filename = f[2][1:-1]
        self.filename = filename
        self.linenroffset = self._scanner.get_line_number()-linenr+1

        
# _SLfilter
class _SLfilter:
    """Used by the sltokenizer to filter the Shading Language source.

    Only the shader and function definitions remain, the bodies will
    be dropped. The filtered result will be used as input for the
    actual parser.
    """
    def __init__(self):
        # Current {}-depth (0 = outside any {..})
        self.curly_depth = 0
        # Current ()-depth
        self.bracket_depth = 0
        # Receives the source code that'll get passed to the parser
        self.SLsource = ""
        self.stream_enabled = True

        self.current_filename = None
        
    def eater(self, type, tok, start, end, line, filename):
        """Record only tokens with depth 0."""
        if filename!=self.current_filename:
            self.current_filename = filename
            if self.SLsource!="" and self.SLsource[-1]!="\n":
                self.SLsource+="\n"
            self.SLsource+='# %d "%s"\n'%(start[0],filename)
        
        if tok=="}":
            self.curly_depth-=1
            if self.curly_depth==0 and self.bracket_depth==0:
                self.stream_enabled = True
        elif tok==")":
            self.bracket_depth-=1

        # Always record newlines so that line numbers won't get messed up
        if self.stream_enabled or tok=="\n":
            self.SLsource+=tok
            
        if tok=="{":
            if self.curly_depth==0 and self.bracket_depth==0:
                self.stream_enabled = False
            self.curly_depth+=1
        elif tok=="(":
            self.bracket_depth+=1


# slparams
def slparams(slfile=None, cpp=None, cpperrstream=sys.stderr, slname=None, includedirs=None, defines=None):
    """Extracts the shader parameters from a RenderMan Shader file.

    The argument *slfile* is either the name of a compiled shader, the name of
    the shader source file (``*.sl``) or a file-like object that provides the
    shader sources.
    
    *cpp* determines how the shader source is preprocessed.
    It can either be a string containing the name of an external
    preprocessor tool (such as ``cpp``) that must take the file name as
    parameter and dump the preprocessed output to stdout or it can be
    a callable that takes *slfile* and *cpperrstream* as input and returns
    the preprocessed sources as a string. If the external
    preprocessor does not produce any data a :exc:`PreprocessorNotFound`
    exception is thrown.
    The error stream of the preprocessor is written to the object
    specified by *cpperrstream* which must have a :meth:`write()`
    method. If *cpperrstream* is ``None``, the error stream is ignored.

    If *cpp* is ``None`` a simple internal preprocessor based on the
    :mod:`simplecpp` module is used.
    The *slname* argument is an alias for *slfile*, it is only available
    for backwards compatibility.
    
    *includedirs* is a list of strings that contain directories where to
    look for include files. *defines* is a list of tuples (*name*, *value*)
    that specify the predefined symbols to use.
    
    The function returns a list of shader info objects. These objects have
    four attributes:
    
    - ``type``: The type of the shader (surface, displacement, etc.)
    - ``name``: The name of the shader
    - ``params``: The shader parameters (see below)
    - ``meta``: The shader meta data. How exactly meta data is specified depends
      on the renderer you are using.
     
    The parameters are given as a list of shader parameter objects
    describing each parameter. A shader parameter object has the
    following attributes:

    - ``outputSpec``: The output specifier (either ``"output"`` or an empty string)
    - ``storage``: The storage class (``"uniform"`` or ``"varying"``)
    - ``type``: The parameter type
    - ``size``: The array length or ``None`` if the parameter is not an array
    - ``name``: The name of the parameter
    - ``space``: The space in which a point-like type was defined
    - ``default``: The "raw" default value. If the input was a shader source file,
      this will always be a string containing an expression. If the input was
      a compiled shader this will already be an appropriate Python value.
      You should never use this value directly, but always use :func:`convertdefault()`
      to obtain a value which can be further processed. This way, your code
      will work for both, compiled shaders and shader source files.
    
    For backwards compatibility, the shader info object behaves like a
    3-tuple (*type*, *name*, *params*). The meta data can only be accessed via
    name though. The shader parameter objects can also be used like 7-tuples
    containing the above data (in the order given above).
    
    Example (output slightly reformatted for better readability)::

      >>> from cgkit import slparams
      >>> shaders = lparams.slparams("plastic.sl")
      >>> print shaders
      [('surface', 'plastic', 
        [('', 'uniform', 'float', None, 'Ka', None, '1'),
         ('', 'uniform', 'float', None, 'Kd', None, '0.5'),
         ('', 'uniform', 'float', None, 'Ks', None, '0.5'),
         ('', 'uniform', 'float', None, 'roughness', None, '0.1'),
         ('', 'uniform', 'color', None, 'specularcolor', 'rgb', '1')])]
      >>> shaders[0].type
      'surface'
      >>> shaders[0].name
      'plastic'
      >>> for param in shaders[0].params: print param.name
      ... 
      Ka
      Kd
      Ks
      roughness
      specularcolor
      >>> shaders[0].meta
      {}

    The parser used inside this function was generated using the parser generator
    `Yapps <http://theory.stanford.edu/~amitp/Yapps/>`_  by Amit Patel.
    """

    if slname is not None:
        slfile = slname
    if slfile is None:
        return []
    
    # Check if the input file is a string referring to a compiled shader
    # (suffix != .sl). If so, use the sloargs module to get the shader information
    if isinstance(slfile, str):
        if os.path.splitext(slfile)[1].lower()!=".sl":
            if (_has_sloargs):
                return sloargs.slparams(slfile)
            else:
                raise _sloargs_importerror

    # Run the preprocessor on the input file...
    
    slsrc = preprocess(cpp, slfile, cpperrstream=cpperrstream, defines=defines, includedirs=includedirs)
    f = io.StringIO(slsrc)
    
    # ...and filter it, so that only the shader and function
    # definitions remain...
    filter = _SLfilter()
    sltokenize.tokenize(f.readline, filter.eater)
    f.close()

#    print filter.SLsource

    # Parse the filtered source code...
    scanner = _slparser._SLParserBaseScanner(filter.SLsource)
    parser = _SLParser(scanner)
    
#    return wrap_error_reporter(parser, "definitions")

    try:
        lst = getattr(parser, "definitions")()
        # Turn the 3-tuples into _ShaderInfo objects
        return [_ShaderInfo(*tup) for tup in lst]
    except _slparser.NoMoreTokens as err:
        raise NoMoreTokens("No more tokens")
    except _slparser.SyntaxError as err:
        scanner = parser._scanner
        input = scanner.input
        cpplineno = scanner.get_line_number()
        lineno = cpplineno-parser.linenroffset
        # '+1' at the end to exclude the newline. If no newline is found
        # rfind() returns -1, so by adding +1 we get 0 which is what we want.
        start = input.rfind("\n", 0, err.charpos)+1
        end   = input.find("\n", err.charpos, -1)
        origline = input[start:end].replace("\t", " ")
        line = " "+origline.lstrip()
        errpos = err.charpos-start-(len(origline)-len(line))
#        print "%s^"%((errpos)*" ")
        msg = 'Syntax error in "%s", line %d:\n'%(parser.filename, lineno)
        msg += '\n%s\n%s^'%(line, errpos*" ")
        exc = SyntaxError(msg)
        exc.filename = parser.filename
        exc.lineno = lineno
        exc.line = line
        exc.errpos = errpos
        raise exc

# preprocess
def preprocess(cpp, file, cpperrstream=sys.stderr, defines=None, includedirs=None):
    """Preprocess an input file.

    cpp is either a string containing the name of an external preprocessor
    command (e.g. 'cpp'), a callable function that takes file and
    cpperrstream as input and returns the preprocessed source code as
    a string or None in which case the built-in preprocessor is used.
    If the external preprocessor produces no data it is assumed that
    it was not found a a PreprocessorNotFound exception is thrown.

    file is either a string containing the input file name or it's a
    file-like object.
    If the input file doesn't exist, an IOError is thrown.

    cpperrstream is a stream that receives any error messages from
    the preprocessor. If it's None, error messages are ignored.

    defines is a list of tuples (name, value) that specify the predefined
    symbols. includedirs is a list of strings with include paths.

    The function returns the preprocessed sources as a string.
    """

    if cpp==None:
        cpp = simplecpp.PreProcessor(defines=defines, includedirs=includedirs)
        return cpp(file, cpperrstream)
    # Is cpp a callable then invoke it and return the result
    elif isinstance(cpp, collections.Callable):
        return cpp(file, cpperrstream)

    # no callable, so cpp contains the name of an external preprocessor tool

    # Variables:
    #
    # cmd: The command that is used to execute the preprocessor.
    #      It either contains only the cpp name or the cpp name + input file
    # slsrc: (string) Unprocessed SL source if it should be passed via stdin
    #      otherwise it's None.
    # ppslsrc: Preprocessed SL source (this is the result).

    # Is file a string? Then it's a file name...
    if isinstance(file, str):
        cmdtoks = [cpp]
        if defines!=None:
            for name,value in defines:
                if value==None:
                    cmdtoks.append("-D%s"%name)
                else:
                    cmdtoks.append("-D%s=%s"%(name,value))
        if includedirs!=None:
            cmdtoks.extend(["-I%s"%dir for dir in includedirs])
        cmdtoks.append(file)
        cmd = " ".join(cmdtoks)
        print (cmd)
        slsrc = None
        # Check if the file exists and can be accessed (by trying to open it)
        # If the file doesn't exist, an exception is thrown.
        dummy = open(file)
        dummy.close()
    # ...otherwise it's a file-like object
    else:
        cmd = "%s"%cpp
        slsrc = file.read()

    # The preprocessed data will be in slsrc.
    stdin, stdout, stderr = os.popen3(cmd)
    if slsrc!=None:
        stdin.write(slsrc)
    stdin.close()
    ppslsrc = stdout.read()
    stdout.close()
    if cpperrstream!=None:
        errs = stderr.read()
        cpperrstream.write(errs)
    # No data? Then it's assumed that the preprocessor couldn't be found
    if len(ppslsrc)==0:
        raise PreprocessorNotFound("Calling '%s' didn't produce any data."%cmd)

    return ppslsrc


# Setup local namespace for convertdefault()
_local_namespace = {}
exec ("from cgkit.sl import *", _local_namespace)
   
def convertdefault(paramtuple):
    """Converts the default value of a shader parameter into a Python type.

    *paramtuple* must be a 7-tuple (or parameter object) as returned by 
    :func:`slparams()`.
    The function returns a Python object that corresponds to the default
    value of the parameter. If the default value can't be converted
    then ``None`` is returned. Only the functions that are present in the
    :mod:`sl<cgkit.sl>` module are evaluated. If a default value calls a user defined
    function then ``None`` is returned.

    The SL types will be converted into the following Python types:

        +------------+-------------+
        | SL type    | Python type |
        +============+=============+
        | ``float``  | ``float``   |
        +------------+-------------+
        | ``string`` | ``string``  |
        +------------+-------------+
        | ``color``  | ``vec3``    |
        +------------+-------------+
        | ``point``  | ``vec3``    |
        +------------+-------------+
        | ``vector`` | ``vec3``    |
        +------------+-------------+
        | ``normal`` | ``vec3``    |
        +------------+-------------+
        | ``matrix`` | ``mat4``    |
        +------------+-------------+

    Arrays will be converted into lists of the corresponding type.
    """
    global _local_namespace
    
    typ = paramtuple[2]
    arraylen = paramtuple[3]
    defstr = paramtuple[6]
    
    # The default value is not a string? Then it already contains the
    # converted default value (this is the case when the value was
    # extracted from a compiled shader).
    if not isinstance(defstr, str):
        # Make sure that point-like types are returned as vec3 and matrix types
        # are returned as mat4.
        if typ in ["color", "point", "vector", "normal"]:
            retType = cgtypes.vec3
        elif typ=="matrix":
            retType = cgtypes.mat4
        else:
            # No vec3/mat4 type, then just return the value
            return defstr
        # Cast the value...
        if arraylen is None:
            return retType(defstr)
        else:
            return [retType(v) for v in defstr]

    # Replace {} with [] so that SL arrays look like Python lists
    if arraylen is not None:
        defstr = defstr.replace("{","[").replace("}","]")
    # If the parameter is not an array, then create an array with one
    # element (to unify further processing). It will be unwrapped in the end
    if arraylen is None:
        defstr = "[%s]"%defstr
    # Evaluate the string to create "raw" Python types (lists and tuples)
    try:
        rawres = eval(defstr, globals(), _local_namespace)
    except:
        return None

    # Convert into the appropriate type...
    if typ=="float":
        try:
            res = [float(x) for x in rawres]
        except:
            return None
    elif typ=="color" or typ=="point" or typ=="vector" or typ=="normal":
        try:
            res = [cgtypes.vec3(x) for x in rawres]
        except:
            return None
    elif typ=="matrix":
        try:
            res = [cgtypes.mat4(x) for x in rawres]
        except:
            return None
    elif typ=="string":
        try:
            res = [str(x) for x in rawres]
        except:
            return None

    if arraylen is None:
        if len(res)==0:
            return None
        else:
            res = res[0]

    return res

######################################################################

if __name__=="__main__":
    pass



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: riutil.py,v 1.1.1.1 2004/12/12 14:31:21 mbaas Exp $

import os.path
import ctypes
import functools
from . import cgtypes
from . import rmanlibutil
from . import _pointcloud
try:
    import numpy
    _numpy_available = True
except ImportError:
    _numpy_available = False

def _arrayPointer(a, n):
    """Check an array and return a pointer to its start.
    
    a is an array object and n the minimum number of values the array
    must contain. An exception is thrown when the array is too small
    or its type is not float (4 bytes) or when a is not a recognized
    array at all.
    """
    # Is the array a ctypes array?
    if isinstance(a, ctypes.Array):
        if a._type_!=ctypes.c_float:
            raise TypeError("Float array expected")
        if len(a)<n:
            raise TypeError("Array is not large enough")
        return ctypes.addressof(a)
    # Does the array support the array interface? (e.g. a numpy array)
    elif hasattr(a, "__array_interface__"):
        strides = a.__array_interface__.get("strides", None)
        typestr = a.__array_interface__.get("typestr", None)
        shape = a.__array_interface__.get("shape", None)
        data = a.__array_interface__.get("data", None)
        if strides is not None:
            raise TypeError("Unsupported array type (strides are not supported)")
        if not (isinstance(typestr, str) and len(typestr)==3 and typestr[1:3]=="f4"):
            raise TypeError("Unsupported array type (the array must contain 4-byte floats)")
        if type(shape) is not tuple:
            raise TypeError("Unsupported array type (unexpected shape value)")
        numFloats = functools.reduce(lambda x,y: x*y, shape)
        if numFloats<n:
            raise TypeError("Array is not large enough")
        if type(data) is not tuple or len(data)!=2:
            raise TypeError("Unsupported array type (unsupported data value)")
        if data[1]==True:
            raise TypeError("Array is read-only")
        if type(data[0]) not in [int, int]:
            raise TypeError("Unsupported array type (data pointer is not an int)")
        return data[0]
    # Unknown array
    else:
        raise TypeError("Unknown array type")


class PtcReader(object):
    """Point cloud reader class.
    
    An instance of this class is returned by the open() function.
    """
    
    def __init__(self, fileName, libName):
        """Constructor.
        
        fileName is the name of the point cloud file and libName the name
        of the shared library that implements the point cloud API.
        """
        object.__init__(self)
        
        self._handle = None
        
        ptclib = self._loadPtcLib(libName)
        
        # Store the functions pointers that are still required
        self._PtcReadDataPoint = ptclib.PtcReadDataPoint
        self._PtcClosePointCloudFile = ptclib.PtcClosePointCloudFile

        self._PtcGetPointCloudInfo = ptclib.PtcGetPointCloudInfo

        # Set 64 as default (which is the maximum in PRMan (when using this API call))
        nvars = ctypes.c_int(64)

        fileNameEncoded = fileName.encode("ascii")

        # Just open the file to find out the number of variables in the file...
        # (3Delight only)
        if "3delight" in os.path.basename(libName):
            handle = ptclib.PtcOpenPointCloudFile(fileNameEncoded, ctypes.byref(nvars), None, None)
            if handle is None:
                raise IOError("Cannot open point cloud file %s"%fileName)
            ptclib.PtcClosePointCloudFile(handle)

        # Now prepare storage for the variable names and types and open the file for real...
        numVars = nvars.value
        types = (numVars*ctypes.c_char_p)()
        names = (numVars*ctypes.c_char_p)()
        handle = ptclib.PtcOpenPointCloudFile(fileNameEncoded, ctypes.byref(nvars), types, names)
        if handle is None:
            raise IOError("Cannot open point cloud file %s"%fileName)
        
        self._handle = handle

        self.name = fileName
        # The dictionary containing the point cloud attributes.
        # Access to the attributes is provided via properties.
        self._ptcAttrs = {}
        
        code = ""
        idx = 0
        vars = []
        for i in range(nvars.value):
            name = names[i].decode("ascii")
            type = types[i].decode("ascii")
            vars.append((type, name))
            if type=="float":
                code += "dataDict['%s'] = data[%s]\n"%(name,idx)
                idx += 1
            elif type in ["vector", "point", "normal", "color"]:
                code += "dataDict['%s'] = data[%s:%s]\n"%(name,idx,idx+3)
                idx += 3
            elif type=="matrix":
                code += "dataDict['%s'] = data[%s:%s]\n"%(name,idx,idx+16)
                idx += 16
            else:
                raise RuntimeError("Unknown variable type in point cloud file: %s"%type)
        # Compile the code that will pick the correct data components and put them into a dict
        self._dataCollectionCode = compile(code, "<string>", "exec")
        
        self._ptcAttrs["variables"] = vars
        
        # Get the npoints attribute...
        n = ctypes.c_int()
        ptclib.PtcGetPointCloudInfo.argtypes = [ptclib.PtcPointCloud, ctypes.c_char_p, ctypes.POINTER(ctypes.c_int)]
        if ptclib.PtcGetPointCloudInfo(handle, b"npoints", ctypes.byref(n))==1:
            self._ptcAttrs["npoints"] = n.value
        # Get the datasize attribute
        if ptclib.PtcGetPointCloudInfo(handle, b"datasize", ctypes.byref(n))==1:
            self._ptcAttrs["datasize"] = n.value
        # Get the bbox attribute
        ptclib.PtcGetPointCloudInfo.argtypes = [ptclib.PtcPointCloud, ctypes.c_char_p, ctypes.POINTER(ctypes.c_float)]
        b = (6*ctypes.c_float)()
        if ptclib.PtcGetPointCloudInfo(handle, b"bbox", b)==1:
            self._ptcAttrs["bbox"] = list(b)
        # Get the format attribute
        f = (3*ctypes.c_float)()
        if ptclib.PtcGetPointCloudInfo(handle, b"format", f)==1:
            self._ptcAttrs["format"] = tuple(f)
        # Get the world2eye attribute
        m = (16*ctypes.c_float)()
        if ptclib.PtcGetPointCloudInfo(handle, b"world2eye", m)==1:
            self._ptcAttrs["world2eye"] = list(m)
        # Get the world2ndc attribute
        if ptclib.PtcGetPointCloudInfo(handle, b"world2ndc", m)==1:
            self._ptcAttrs["world2ndc"] = list(m)

        if self.npoints is None:
            raise IOError("Could not obtain the number of points in point cloud file %s."%fileName)
        if self.npoints<0:
            raise ValueError("The number of points in the point cloud file is negative (%s)."%self.npoints)
        if self.datasize is None:
            raise IOError("Could not obtain datasize value from point cloud file %s."%fileName)

        # The number of points that can still be read before eof is hit
        self._numPointsLeft = self.npoints

        # Set up storage for reading individual data points
        self._pos = (3*ctypes.c_float)()
        self._normal = (3*ctypes.c_float)()
        self._radius = ctypes.c_float()
        self._data = (self.datasize*ctypes.c_float)()

    def __del__(self):
        """Destructor.
        
        Closes the file.
        """
        self.close()

    @property
    def variables(self):
        """Return the variables attribute."""
        return self._ptcAttrs.get("variables", None)
    
    @property
    def npoints(self):
        """Return the npoints attribute."""
        return self._ptcAttrs.get("npoints", None)

    @property
    def datasize(self):
        """Return the datasize attribute."""
        return self._ptcAttrs.get("datasize", None)

    @property
    def bbox(self):
        """Return the bbox attribute."""
        return self._ptcAttrs.get("bbox", None)

    @property
    def format(self):
        """Return the format attribute."""
        return self._ptcAttrs.get("format", None)

    @property
    def world2eye(self):
        """Return the world2eye attribute."""
        return self._ptcAttrs.get("world2eye", None)

    @property
    def world2ndc(self):
        """Return the world2ndc attribute."""
        return self._ptcAttrs.get("world2ndc", None)
    
    def iterAttrs(self):
        """Iterate over all attributes defined in the file.
        
        Yields tuples (name,value).
        """
        return list(self._ptcAttrs.items())
    
    def close(self):
        """Close the point cloud file.
        
        This method is also called from the destructor.
        """
        if self._handle is not None:
            self._PtcClosePointCloudFile(self._handle)
            self._handle = None
    
    def readDataPoint(self):
        """Read the next data point.
        
        Returns a tuple (pos, normal, radius, dataDict) where pos and normal
        are 3-tuples of floats, radius is a single float and dataDict a
        dictionary with the extra variables that are attached to the point.
        If no more point is available an EOFError exception is thrown.
        An IOErrror exception is thrown when an error occurs during reading or
        when the file has already been closed.
        """
        if self._handle is None:
            raise IOError("The point cloud file has already been closed (%s)"%self.name)
        if self._numPointsLeft==0:
            raise EOFError("There are no more points left to read from point cloud file %s"%self.name)
        
        self._numPointsLeft -= 1
        res = self._PtcReadDataPoint(self._handle, self._pos, self._normal, self._radius, self._data)
        if res==0:
            raise IOError("Error while reading data point from point cloud file %s"%self.name)
        else:
            dataDict = {}
            data = self._data
            exec(self._dataCollectionCode)
            return tuple(self._pos), tuple(self._normal), self._radius.value, dataDict

    def readDataPoints(self, numPoints, buffer):
        """Read a sequence of data points.
        
        numPoints is the number of points to read. buffer is either a single
        buffer that will receive all values or a tuple (pointbuf, normalbuf,
        radiusbuf, databuf) that contains the individual buffers for the
        respective values. A buffer must always be large enough to hold
        numPoints values.
        The function accepts ctypes arrays as buffers or any sequence object
        that supports the array interface (such as numpy arrays).
        
        The return value is the number of points that have actually
        been read (additional items in the buffers remain at their previous
        value). When 0 is returned, the end of the file has been reached.
        """
        if numPoints<=0:
            return 0
        
        # Are there 4 individual buffers?
        if type(buffer) is tuple:
            if len(buffer)!=4:
                raise ValueError("Expected four individual buffers, but got %s"%len(buffer))
            pbuf,nbuf,rbuf,dbuf = buffer
            pntPtr = _arrayPointer(pbuf, 3*numPoints)
            normPtr = _arrayPointer(nbuf, 3*numPoints)
            radPtr = _arrayPointer(rbuf, numPoints)
            dataPtr = _arrayPointer(dbuf, self.datasize*numPoints)
            pntStride = 3
            normStride = 3
            radStride = 1
            dataStride = self.datasize
        # There is only one single buffer for all values
        else:
            sizeOfFloat = 4
            pntStride = 7+self.datasize
            normStride = pntStride
            radStride = pntStride
            dataStride = pntStride
            pntPtr = _arrayPointer(buffer, pntStride*numPoints)
            normPtr = pntPtr+3*sizeOfFloat
            radPtr = normPtr+3*sizeOfFloat
            dataPtr = radPtr+sizeOfFloat
        
        num = min(numPoints, self._numPointsLeft)
        
        # Read the points
        self._numPointsLeft -= num
        _pointcloud.readDataPoints(ctypes.addressof(self._PtcReadDataPoint), self._handle, num,
                                   pntPtr, pntStride, normPtr, normStride, radPtr, radStride, dataPtr, dataStride)
        return num

    def iterPoints(self):
        """Iterate over all the points in the file.
        
        Yields tuples (point,normal,radius,data) for every point in the file.
        This is equivalent to calling readDataPoint() repeatedly.
        """
        while self._numPointsLeft>0:
            yield self.readDataPoint()
            
    def iterBatches(self, batchSize=1000, combinedBuffer=False, numpyArray=False):
        """Iterate over point batches.
        
        Reads batchSize points at once and yields one or more buffers
        containing the data.
        combinedBuffer determines whether all data is written into one single
        buffer or if there is an individual buffer for the point, normal, radius
        and data.
        numpyArray determines whether the buffers are created as numpy arrays
        or ctypes arrays.
        """
        global _numpy_available
        
        if batchSize<1:
            raise ValueError("Invalid point batch size: %s"%batchSize)
        if numpyArray and not _numpy_available:
            raise ImportError("numpy is not available") 
        
        num = self._numPointsLeft
        buffer = None
        bufLen = 0
        while num>0:
            n = min(batchSize, num)
            if bufLen!=n:
                if combinedBuffer:
                    if numpyArray:
                        buffer = numpy.zeros(shape=(n,7+self.datasize), dtype=numpy.float32)
                    else:
                        buffer = (((7+self.datasize)*n)*ctypes.c_float)()
                else:
                    if numpyArray:
                        ps = numpy.zeros(shape=(n,3), dtype=numpy.float32)
                        ns = numpy.zeros(shape=(n,3), dtype=numpy.float32)
                        rs = numpy.zeros(shape=(n,), dtype=numpy.float32)
                        ds = numpy.zeros(shape=(n,self.datasize), dtype=numpy.float32)
                    else:
                        ps = (3*n*ctypes.c_float)()
                        ns = (3*n*ctypes.c_float)()
                        rs = (n*ctypes.c_float)()
                        ds = (self.datasize*n*ctypes.c_float)()
                    buffer = (ps,ns,rs,ds)
                bufLen = n
            self.readDataPoints(n, buffer)
            yield buffer
            num -= n

    def _loadPtcLib(self, libName):
        """Load a RenderMan library providing the point cloud API.
        """
        resolvedLibName = rmanlibutil.resolveRManLib(libName)
        ptclib = ctypes.cdll.LoadLibrary(resolvedLibName)
        
        ptclib.PtcPointCloud = ctypes.c_void_p
        
        # Reading point cloud files
        ptclib.PtcOpenPointCloudFile.argtypes = [ctypes.c_char_p,
                                                 ctypes.POINTER(ctypes.c_int),
                                                 ctypes.POINTER(ctypes.c_char_p),
                                                 ctypes.POINTER(ctypes.c_char_p)]
        ptclib.PtcOpenPointCloudFile.restype = ptclib.PtcPointCloud
        
        ptclib.PtcGetPointCloudInfo.argtypes = [ptclib.PtcPointCloud, ctypes.c_char_p, ctypes.c_char_p]
        ptclib.PtcGetPointCloudInfo.restype = ctypes.c_int
        
        ptclib.PtcReadDataPoint.argtypes = [ptclib.PtcPointCloud,
                                            ctypes.POINTER(ctypes.c_float),
                                            ctypes.POINTER(ctypes.c_float),
                                            ctypes.POINTER(ctypes.c_float),
                                            ctypes.POINTER(ctypes.c_float)]
        ptclib.PtcReadDataPoint.restype = ctypes.c_int
        
        ptclib.PtcClosePointCloudFile.argtypes = [ptclib.PtcPointCloud]
        ptclib.PtcClosePointCloudFile.restype = None
        
        return ptclib
    
    
class PtcWriter:
    """Point cloud writer class.
    
    An instance of this class is returned by the open() function.
    """
    
    def __init__(self, fileName, libName, vars, world2eye, world2ndc, format):
        """Constructor.
        
        fileName is the name of the point cloud file and libName the name
        of the shared library that implements the point cloud API.
        vars is a list of tuples (type, name) that specifies what additional
        variables to write into the file. world2eye and world2ndc are 4x4
        matrices and format a tuple
        """
        
        self._handle = None
        
        ptclib = self._loadPtcLib(libName)
        
        # Store the functions pointers that are still required
        self._PtcWriteDataPoint = ptclib.PtcWriteDataPoint
        self._PtcFinishPointCloudFile = ptclib.PtcFinishPointCloudFile

        self.name = fileName
        
        xres,yres,aspect = format
        
        w2e = self._matrixToCTypes(world2eye)
        w2n = self._matrixToCTypes(world2ndc)
        
        n = len(vars)
        cvartypes = (n*ctypes.c_char_p)()
        cvarnames = (n*ctypes.c_char_p)()
        idx = 0
        code = ""
        for i in range(n):
            type,name = vars[i]
            cvartypes[i] = type.encode("ascii")
            cvarnames[i] = name.encode("ascii")
            if type=="float":
                code += "cdata[%s] = data.get('%s', 0.0)\n"%(idx, name)
                idx += 1
            elif type in ["vector", "point", "normal", "color"]:
                code += "cdata[%s:%s] = data.get('%s', (0.0,0.0,0.0))\n"%(idx,idx+3,name)
                idx += 3
            elif type=="matrix":
                code += "cdata[%s:%s] = data.get('%s'), (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0))\n"%(idx,idx+16,name)
                idx += 16
            else:
                raise RuntimeError("Unknown point cloud variable type: %s"%type)

        self.datasize = idx
        self._dataInitCode = code
        
        cformat = (3*ctypes.c_float)(float(xres), float(yres), float(aspect))
        
        fileNameEncoded = fileName.encode("ascii")
        handle = ptclib.PtcCreatePointCloudFile(fileNameEncoded, n, cvartypes, cvarnames, w2e, w2n, cformat)
        if handle is None:
            raise IOError("Cannot open point cloud file %s for writing"%fileName)

        self._handle = handle
        
    def _matrixToCTypes(self, m):
        """Convert a matrix into a ctypes array.
        
        m can be any object that contains 16 floats (the values may be nested).
        """
        if isinstance(m, cgtypes.mat4):
            values = m.toList(rowmajor=True)
        else:
            values = []
            for v in m:
                try:
                    values.extend(list(v))
                except:
                    values.append(v)
            if len(values)!=16:
                raise ValueError("Matrix must be composed of 16 values, got %s instead."%len(values))
        return (16*ctypes.c_float)(*values)
        
    def __del__(self):
        self.close()
    
    def close(self):
        """Close the point cloud file.
        
        This method is also called from the destructor.
        """
        if self._handle is not None:
            self._PtcFinishPointCloudFile(self._handle)
            self._handle = None

    def writeDataPoint(self, point, normal, radius, data):
        """Write a point into the point cloud file.
        
        point and normal are vectors (any 3-sequence of floats) and radius
        a float. data is a dict that contains the extra variables that
        must have been declared in the constructor. Undeclared values are
        ignored, missing declared values are set to 0.
        """
        if self._handle is None:
            raise IOError("The point cloud file has already been closed.")

        p = (3*ctypes.c_float)(*tuple(point))
        n = (3*ctypes.c_float)(*tuple(normal))
        cdata = (self.datasize*ctypes.c_float)()
        exec(self._dataInitCode)
        res = self._PtcWriteDataPoint(self._handle, p, n, radius, cdata)
        if res==0:
            raise IOError("Failed to write point cloud data point")

    def writeDataPoints(self, numPoints, buffer):
        """Write a sequence of data points.
        
        numPoints is the number of points to write. buffer is either a single
        buffer that contains all values or a tuple (pointbuf, normalbuf,
        radiusbuf, databuf) that each contains the respective value.
        The buffers must contain at least numPoints items.
        The function accepts ctypes arrays as buffers or any sequence object
        that supports the array interface (such as numpy arrays).
        """
        # Are there 4 individual buffers?
        if type(buffer) is tuple:
            if len(buffer)!=4:
                raise ValueError("Expected four individual buffers, but got %s"%len(buffer))
            pbuf,nbuf,rbuf,dbuf = buffer
            pntPtr = _arrayPointer(pbuf, 3*numPoints)
            normPtr = _arrayPointer(nbuf, 3*numPoints)
            radPtr = _arrayPointer(rbuf, numPoints)
            dataPtr = _arrayPointer(dbuf, self.datasize*numPoints)
            pntStride = 3
            normStride = 3
            radStride = 1
            dataStride = self.datasize
        # There is only one single buffer for all values
        else:
            sizeOfFloat = 4
            pntStride = 7+self.datasize
            normStride = pntStride
            radStride = pntStride
            dataStride = pntStride
            pntPtr = _arrayPointer(buffer, pntStride*numPoints)
            normPtr = pntPtr+3*sizeOfFloat
            radPtr = normPtr+3*sizeOfFloat
            dataPtr = radPtr+sizeOfFloat
        
        # Write the points
        _pointcloud.writeDataPoints(ctypes.addressof(self._PtcWriteDataPoint), self._handle, numPoints,
                                    pntPtr, pntStride, normPtr, normStride, radPtr, radStride, dataPtr, dataStride)

    def _loadPtcLib(self, libName):
        """Load a RenderMan library providing the point cloud API.
        """
        resolvedLibName = rmanlibutil.resolveRManLib(libName)
        ptclib = ctypes.cdll.LoadLibrary(resolvedLibName)
        
        ptclib.PtcPointCloud = ctypes.c_void_p

        # Writing point cloud files
        ptclib.PtcCreatePointCloudFile.argtypes = [ctypes.c_char_p,
                                                   ctypes.c_int,
                                                   ctypes.POINTER(ctypes.c_char_p),
                                                   ctypes.POINTER(ctypes.c_char_p),
                                                   ctypes.POINTER(ctypes.c_float),
                                                   ctypes.POINTER(ctypes.c_float),
                                                   ctypes.POINTER(ctypes.c_float)]
        ptclib.PtcCreatePointCloudFile.restype = ptclib.PtcPointCloud
        
        ptclib.PtcWriteDataPoint.argtypes = [ptclib.PtcPointCloud,
                                             ctypes.POINTER(ctypes.c_float),
                                             ctypes.POINTER(ctypes.c_float),
                                             ctypes.c_float,
                                             ctypes.POINTER(ctypes.c_float)]
        ptclib.PtcWriteDataPoint.restype = ctypes.c_int
        
        ptclib.PtcFinishPointCloudFile.argtypes = [ptclib.PtcPointCloud]
        ptclib.PtcFinishPointCloudFile.restype = None
        
        return ptclib


def open(fileName, mode="r", libName=None, **kwargs):
    """Open a point cloud file for reading or writing.
    
    *fileName* is the name of the point cloud file. *mode* is either ``"r"``
    for reading a file or ``"w"`` for writing a new point cloud file.
    *libName* is the library name that implements the point cloud API.
    When mode is ``"w"``, the following additional keyword arguments must
    be present:
    
    - ``vars``: A list of tuples (*type*, *name*) that defines what additional variables to write
    - ``world2eye``: The world2eye matrix 
    - ``world2ndc``: The world2ndc matrix
    - ``format``: A tuple (*xres*, *yres*, *aspect*)
    
    Depending on the mode, the function either returns a :class:`PtcReader` or
    :class:`PtcWriter` object.
    """
    if mode=="r":
        return PtcReader(fileName, libName, **kwargs)
    elif mode=="w":
        return PtcWriter(fileName, libName=libName, **kwargs)
    else:
        raise ValueError('Invalid file mode: "%s" (expected "r" or "w")'%mode)

###################################################################

if __name__=="__main__":
    from cgkit.cgtypes import *
    import random
    
    if 1:
        print("---WRITE---")
        ptc = open("test.ptc", "w", "3delight", vars=[("float", "spam"), ("vector", "dir")], world2eye=None, world2ndc=None, format=(320,240,1.333))
        for i in range(100):
            x = random.random()
            y = random.random()
            z = random.random()
            ptc.writeDataPoint((x,y,z), (0,1,0), 0.2, {"spam":0.5})
        ptc.close()
    
    print("----READ----")
    rd = open("test.ptc", "r", "3delight")
    print((rd.variables))
    print(("npoints:",rd.npoints))
    print(("datasize",rd.datasize))
    print(("bbox",rd.bbox))
    print(("format",rd.format))
    print(("world2eye",rd.world2eye))
    print(("world2ndc",rd.world2ndc))
    p,n,r,data = rd.readDataPoint()
    print((p,n,r,data))
    print((rd.readDataPoint()))



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: component.py,v 1.3 2005/08/15 15:47:56 mbaas Exp $

## \file component.py
## Contains the Component base class.

from ._core import Component as _Component
from .Interfaces import ISceneItem
import copy, inspect, re
from . import protocols
from .slots import *
from .globalscene import getScene

# Component
class Component(_Component):
    """Base component class.

    Attributes:

    - name (\c str): The name of the component
    """

    protocols.advise(instancesProvide=[ISceneItem])
    
    def __init__(self, name="", auto_insert=True):
        _Component.__init__(self, name)
        
        if auto_insert:
            getScene().insert(self)

    def protocols(self):
        return [ISceneItem]

    def connect(self, srcslot, dstslot):
        dstslot.setController(srcslot)

############

def _parseFuncDecl(funcdecl):
    """Parse a function declaration string.

    The syntax of the string is like a C function declaration:

    <result_type> <func_name>(<parameters>)

    Example: "vec3 f(double x, double x, double z)"
    
    The function name may be left out

    The function returns a tuple (result_type, func_name, params)
    where params is a list of tuples (name, type, default). The name of a
    parameter might also be empty.
    """
    funcdecl = funcdecl.strip()

    # Parse the result type...
    m = re.match("[a-zA-Z0-9]+ ", funcdecl)
    if m==None:
        raise SyntaxError("No valid return type found")
        
    restype = funcdecl[m.start():m.end()].strip()

    # Parse the function name...
    s = funcdecl[m.end():]
    i = s.find("(")
    if i==-1:
        raise SyntaxError("Parameter block is missing")
    funcname = s[:i].strip()

    # Parse the arguments...
    args = s[i+1:-1].split(",")
    inputs = []
    for s in args:
        s = s.strip()
        f = s.split()
        if len(f)==1:
            type = f[0]
            name = ""
        elif len(f)==2:
            type,name = f
        else:
            raise SyntaxError("Invalid parameter declaration: '%s'"%s)
        inputs.append((name,type,None))

    return restype, funcname, inputs
    
def _inspectFuncDecl(func):
    """Determine the function declaration from the function itself.

    The function returns a tuple (result_type, func_name, params)
    where params is a list of tuples (name, type, default).

    result_type and params may also be None if they cannot be determined
    (i.e. if not every argument to the function has a default value).
    """
    # Extract function name...
    funcname = func.__name__

    args,va,vkw,defaults = inspect.getargspec(func)
    # Only proceed if every argument has a default value
    if defaults==None or len(args)!=len(defaults):
        return None, funcname, None

    # Extract the argument types (=the types of the default values)
    inputs = []
    for name, value in zip(args, defaults):
        type = value.__class__.__name__
        if type=="float":
            type = "double"
        inputs.append((name, type, value))

    # Determine the result type
    v = func()
    restype = v.__class__.__name__
    if restype=="float":
        restype = "double"
    
    return restype, funcname, inputs
        

def _defaultValueLiteral(type, default):
    """Convert a value into a string so that it can be executed again.
    """
    if default==None:
        return ""
    
    if type in ["vec3", "vec4"]:
        return "%s(%s)"%(type,default)
    else:
        return "%s"%default


def createFunctionComponentSource(clsname, restype, funcname, inputs):
    """Create a string that creates a component class.

    """

    res = """from cgkit import _core
from cgkit.globalscene import getScene
    
class %s(Component):
    
    def __init__(self):
        Component.__init__(self, func.__name__)
"""%(clsname)

    # Create output slot
    outname = "output"
    slotclassname = "Procedural%sSlot"%restype.capitalize()
    slotname = "%s_slot"%outname
    res += "        self.%s = %s(self.compute%s)\n"%(slotname, slotclassname, outname.capitalize())
    res += '        self.addSlot("%s", self.%s)\n'%(outname, slotname)

    # Create input slots
    for name, type, default in inputs:
        slotclassname = "%sSlot"%type.capitalize()
        slotname = "%s_slot"%name
        res += "        self.%s = %s(%s)\n"%(slotname, slotclassname, _defaultValueLiteral(type, default))
        res += '        self.addSlot("%s", self.%s)\n'%(name, slotname)
        res += "        self.%s.addDependent(self.%s_slot)\n"%(slotname, outname)
        if name=="time" and type=="double":
            res += "        getScene().timer().time_slot.connect(self.%s)\n"%slotname

#    res += "        self._func_obj = %s\n"%funcname
    res += "        self._func_obj = func\n"

    res += '\n    exec(slotPropertyCode("%s"))\n'%(outname)
    for name,type,default in inputs:
        res += '    exec(slotPropertyCode("%s"))\n'%(name)

    res += """\n    def compute%s(self):
        return self._func_obj("""%(outname.capitalize())

    inames = ["%s=self.%s"%(x[0],x[0]) for x in inputs]
    res += ", ".join(inames)
    res += ")\n"
    
    return res


# createFunctionComponent
def createFunctionComponent(func, funcdeclaration=None):

    if funcdeclaration==None:
        restype, funcname, inputs = _inspectFuncDecl(func)
        if restype==None:
            raise ValueError("Cannot determine the types of the arguments and the return value.")
    else:
        restype, funcname, inputs = _parseFuncDecl(funcdeclaration)

    if funcname==None or funcname=="":
        funcname = func.__name__
    
    clsname = funcname.capitalize()
    s = createFunctionComponentSource(clsname, restype, funcname, inputs)
    ns = copy.copy(globals())
    ns["func"]=func
    exec(s, ns)
    return ns[clsname]
    



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: ifsimport.py,v 1.1.1.1 2004/12/12 14:31:04 mbaas Exp $

import os.path, struct
from .cgtypes import *
from .worldobject import WorldObject
from .trimesh import TriMesh
from .trimeshgeom import TriMeshGeom
from . import pluginmanager

# IfsImporter
class IfsImporter:
    """IFS importer.

    This class imports models from the Brown Mesh Set library which are
    stored in the Indexed Face Set (IFS) format.
    """

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["ifs"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "Indexed face set"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename):
        """Import an IFS file."""

        f = file(filename, "rb")

        # Check the header...
        s = self.readString(f)
        if s!="IFS":
            raise ValueError('The file "%s" is is not a IFS file.'%filename)

        # Read (and ignore) the version number
        s = f.read(4)
        ver = struct.unpack("<f", s)[0]

        # Read the model name
        modelname = self.readString(f)

        # Create the mesh geom
        tm = TriMeshGeom()

        # Read vertices...
        s = self.readString(f)
        if s!="VERTICES":
            raise ValueError("Vertices chunk expected, got '%s' instead."%s)

        s = f.read(4)
        numverts = int(struct.unpack("<I", s)[0])
        tm.verts.resize(numverts)

        for i in range(numverts):
            s = f.read(12)
            x,y,z = struct.unpack("<fff", s)
            tm.verts[i] = vec3(x,y,z)

        # Read faces...
        s = self.readString(f)
        if s!="TRIANGLES":
            raise ValueError("Triangle chunk expected, got '%s' instead."%s)
            
        s = f.read(4)
        numfaces = int(struct.unpack("<I", s)[0])
        tm.faces.resize(numfaces)

        for i in range(numfaces):
            s = f.read(12)
            a,b,c = struct.unpack("<III", s)
            tm.faces[i] = (int(a), int(b), int(c))

        # Create a world object
        obj = TriMesh(name=modelname)
        obj.geom = tm


    def readString(self, fhandle):
        """Read a string.

        \param fhandle Open file handle
        \return String
        """
        s = fhandle.read(4)
        w = int(struct.unpack("<I", s)[0])
        s = fhandle.read(w)
        # Return the string without the trailing \000
        return s[:-1]


######################################################################

# Register the IfsImporter class as a plugin class
pluginmanager.register(IfsImporter)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: call.py,v 1.1.1.1 2004/12/12 14:31:33 mbaas Exp $

## \file call.py
## Contains the Call class.

import string, types


# Call
class Call:
    """Call adapter.

    This class can either be used to wrap a string or a callable
    object with a fixed set of arguments (positional and keyword
    arguments).  When the %Call object is called (without any
    arguments) it executes the string or calls the wrapped function
    with the stored arguments.

    Objects of this class can be compared using == or != and they're
    hashable (so can be used in dictionaries, etc.)

    The purpose of this class is to provide an adapter for GUI commands
    which take no arguments.

    Example:

    \code

    def foo(a, b, c=0):
        print "a:%s  b:%s  c:%s"%(a,b,c)

    >>> f = Call(foo, 1, 2, c=3)
    
    # Call foo with the specified arguments
    >>> f()
    a:1  b:2  c:3
    
    # Print a readable description of the command object
    >>> print f
    foo(1, 2, c=3)

    \endcode

    """
    
    def __init__(self, func, *args, **keyargs):
        """Constructor.

        You must provide the callable object as first argument. All
        additional arguments will be passed to the callable object
        when its called.

        If \a func is a string then there may be one additional argument
        that specifies a dictionary which should be used as global
        (and local) namespace. The dictionary may be specified as positional
        argument following the string or as keyword argument "globals".

        The constructor sets the variable __doc__ to a reasonable value.
        It'll be either the function doc string if there is one or the
        command string itself.

        \param func (\c callable or \c str) The function that should be called
        \param *args  The positional arguments (default = no args)
        \param **keyargs  The keyword arguments (default = no keyword args)
        """
        if isinstance(func, (str,)):
            self._func = func
            self._args = None
            self._keyargs = None
            compilestr = func
            if "globals" in keyargs:
                self._globals = keyargs["globals"]
            elif len(args)==1:
                self._globals = args[0]
            self.__doc__ = None
        elif callable(func):
            self._func = func
            self._args = args
            self._keyargs = keyargs
            compilestr = "apply(self._func, self._args, self._keyargs)"
            self.__doc__ = getattr(func, "__doc__", None)
        else:
            raise TypeError("The 'func' argument must be a string or a callable object.")

        self._code = compile(compilestr, "<command: '%s'>"%str(self), "exec")
        if self.__doc__==None or self.__doc__=="":
            self.__doc__ = str(self)

    def __str__(self):
        if callable(self._func):
            res = getattr(self._func, "__name__", "<unknown>")
            args = [str(x) for x in self._args]
            args += ["%s=%s"%(x,self._keyargs[x]) for x in self._keyargs]
            res = "%s(%s)"%(res,string.join(args,", "))
        else:
            res = self._func
        return res

    def __repr__(self):
        return "<Call cmd='%s'>"%str(self)

    def __call__(self):
        if hasattr(self, "_globals"):
            exec (self._code, self._globals)
        else:
            exec (self._code)

    def __eq__(self, other):
        if not isinstance(other, Call):
            return False

        return (self._func==other._func and
                self._args==other._args and
                self._keyargs==other._keyargs)

    def __ne__(self, other):
        if not isinstance(other, Call):
            return True

        return (self._func!=other._func or
                self._args!=other._args or
                self._keyargs!=other._keyargs)

    def __hash__(self):
        return hash(self._code)
#        return hash(self._func) ^ hash(self._args)


